from ..fragments import FRAGMENT_PRODUCT

MUTATION_CREATE_PRODUCT = '''
mutation CreateProduct($input: CreateProductInput!) {
  createProduct(input: $input) {
    ok
    product {
      ...fragmentProduct
    }
  }
}
''' + FRAGMENT_PRODUCT



"""copy data in web_config

Revision ID: 1f86dd56d829
Revises: 3746a752f344
Create Date: 2021-10-27 16:22:41.444462

"""
from alembic import op

from sqlalchemy.orm.session import Session

from sqlalchemy import MetaData
from flask_sqlalchemy import SQLAlchemy

# revision identifiers, used by Alembic.
revision = "1f86dd56d829"
down_revision = "3746a752f344"
branch_labels = None
depends_on = None


#
convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    # "ck": "ck_%(table_name)s_%(constraint_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = MetaData(naming_convention=convention)

sa = SQLAlchemy(metadata=metadata)


class WebConfig(sa.Model):
    __tablename__ = "web_config"
    id_ = sa.Column(sa.Integer(), primary_key=True)
    json = sa.Column(sa.Text())


class WebConfigOld(sa.Model):
    __tablename__ = "web_config_old"
    config_id = sa.Column(sa.Integer(), primary_key=True)
    head_title = sa.Column(sa.Text())
    toolbar_title = sa.Column(sa.Text())
    devtool_loadingstate = sa.Column(sa.Boolean())
    product_creation_dialog = sa.Column(sa.Boolean())
    product_update_dialog = sa.Column(sa.Boolean())
    product_deletion_dialog = sa.Column(sa.Boolean())


def upgrade():
    import json

    session = Session(bind=op.get_bind())

    with session.no_autoflush:
        olds = session.query(WebConfigOld).all()

        for o in olds:
            config_json = json.dumps(
                {
                    "headTitle": o.head_title,
                    "toolbarTitle": o.toolbar_title,
                    "devtoolLoadingstate": o.devtool_loadingstate,
                    "productCreationDialog": o.product_deletion_dialog,
                    "productUpdateDialog": o.product_update_dialog,
                    "productDeletionDialog": o.product_deletion_dialog,
                }
            )
            n = WebConfig(id_=o.config_id, json=config_json)
            session.add(n)
    session.commit()


def downgrade():
    session = Session(bind=op.get_bind())
    session.query(WebConfig).delete()
    session.commit()



import pytest
import unittest.mock as mock

from acondbs.github.call import exchange_code_for_token

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_requests(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.github.call.requests", y)
    yield y

##__________________________________________________________________||
def test_success(mock_requests, snapshot):

    code = 'code-xyz'

    token_url = 'https://github.com/login/oauth/access_token'
    client_id = 'client_id_0123'
    client_secret = 'client_secret_0123'
    redirect_uri = 'http://localhost/auth'

    return_value = {'access_token': 'token-xxx', 'token_type': 'bearer', 'scope': 'user'}
    mock_requests.post().json.return_value = dict(return_value)

    response = exchange_code_for_token(code, token_url, client_id, client_secret, redirect_uri)

    assert return_value == response
    snapshot.assert_match(mock_requests.post.call_args_list)

##__________________________________________________________________||
def test_error(mock_requests, snapshot):

    code = 'code-xyz'

    token_url = 'https://github.com/login/oauth/access_token'
    client_id = 'client_id_0123'
    client_secret = 'client_secret_0123'
    redirect_uri = 'http://localhost/auth'

    response = {
        'error': 'bad_verification_code',
        'error_description': 'The code passed is incorrect or expired.',
        'error_uri': 'https://docs.github.com/apps/managing-oauth-apps/troubleshooting-oauth-app-access-token-request-errors/#bad-verification-code'
    }
    mock_requests.post().json.return_value = response

    with pytest.raises(Exception) as e:
        exchange_code_for_token(code, token_url, client_id, client_secret, redirect_uri)

    assert response == e.value.args[0]

    snapshot.assert_match(mock_requests.post.call_args_list)

##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[query] 1'] = {
    'data': {
        'allProductRelations': {
            'edges': [
                {
                    'node': {
                        'other': {
                            'name': 'beam1',
                            'productId': '4'
                        },
                        'relationId': '1',
                        'reverse': {
                            'relationId': '2'
                        },
                        'self_': {
                            'name': 'map1',
                            'productId': '1'
                        },
                        'type_': {
                            'name': 'child',
                            'typeId': '2'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1',
                            'productId': '1'
                        },
                        'relationId': '2',
                        'reverse': {
                            'relationId': '1'
                        },
                        'self_': {
                            'name': 'beam1',
                            'productId': '4'
                        },
                        'type_': {
                            'name': 'parent',
                            'typeId': '1'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2',
                            'productId': '5'
                        },
                        'relationId': '3',
                        'reverse': {
                            'relationId': '4'
                        },
                        'self_': {
                            'name': 'map1',
                            'productId': '1'
                        },
                        'type_': {
                            'name': 'child',
                            'typeId': '2'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'map1',
                            'productId': '1'
                        },
                        'relationId': '4',
                        'reverse': {
                            'relationId': '3'
                        },
                        'self_': {
                            'name': 'beam2',
                            'productId': '5'
                        },
                        'type_': {
                            'name': 'parent',
                            'typeId': '1'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam2',
                            'productId': '5'
                        },
                        'relationId': '5',
                        'reverse': {
                            'relationId': '6'
                        },
                        'self_': {
                            'name': 'beam1',
                            'productId': '4'
                        },
                        'type_': {
                            'name': 'child',
                            'typeId': '2'
                        }
                    }
                },
                {
                    'node': {
                        'other': {
                            'name': 'beam1',
                            'productId': '4'
                        },
                        'relationId': '6',
                        'reverse': {
                            'relationId': '5'
                        },
                        'self_': {
                            'name': 'beam2',
                            'productId': '5'
                        },
                        'type_': {
                            'name': 'parent',
                            'typeId': '1'
                        }
                    }
                }
            ]
        }
    }
}

snapshots['test_schema[total-count] 1'] = {
    'data': {
        'allProductRelations': {
            'totalCount': 6
        }
    }
}



import pytest

from ....funcs import assert_mutation

from ...gql import MUTATION_DELETE_PRODUCT_TYPE

QEURY = """
{
  allProductTypes {
    edges {
      node {
        name
        typeId
      }
    }
  }
}
"""

HEADERS = {
    "Authorization": "Bearer 0fb8c9e16d6f7c4961c4c49212bf197d79f14080"  # dojocat
}


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": MUTATION_DELETE_PRODUCT_TYPE, "variables": {"typeId": 2}},
        {"query": QEURY},
        id="delete",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_success(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):

    success = True
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": MUTATION_DELETE_PRODUCT_TYPE, "variables": {"typeId": 12}},
        {"query": QEURY},
        id="error-nonexistent",
    ),
    pytest.param(
        {"query": MUTATION_DELETE_PRODUCT_TYPE, "variables": {"typeId": 1}},
        {"query": QEURY},
        id="error-unempty",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_error(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):
    success = False
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS,
        data_query,
        HEADERS,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||



import pytest

from acondbs.db.sa import sa
from acondbs.models import ProductType, Product, ProductFilePath


##__________________________________________________________________||
@pytest.fixture
def app(app_empty):

    y = app_empty

    # product --- path1
    #          |
    #          +- path2
    type1 = ProductType(name="type1")
    product = Product(name="product", type_=type1)
    path1 = ProductFilePath(path="/path1")
    path2 = ProductFilePath(path="/path2")
    product.paths = [path1, path2]

    # commit
    with y.app_context():
        sa.session.add(product)
        sa.session.commit()
    yield y


##__________________________________________________________________||
def test_relations(app):

    with app.app_context():
        product = Product.query.filter_by(name="product").one_or_none()
        assert "product" == product.name
        assert 2 == len(product.paths)
        path1, path2 = product.paths
        assert "/path1" == path1.path
        assert "/path2" == path2.path
        assert product is path1.product
        assert product is path2.product


def test_delete_product(app):

    with app.app_context():
        product = Product.query.filter_by(name="product").one_or_none()
        sa.session.delete(product)
        sa.session.commit()

    with app.app_context():
        product = Product.query.filter_by(name="product").one_or_none()
        assert product is None
        assert [] == ProductFilePath.query.all()


def test_delete_path(app):

    with app.app_context():
        path1 = ProductFilePath.query.filter_by(path="/path1").one_or_none()
        sa.session.delete(path1)
        sa.session.commit()

    with app.app_context():
        product = Product.query.filter_by(name="product").one_or_none()
        assert product is not None
        path2 = ProductFilePath.query.filter_by(path="/path2").one_or_none()
        assert [path2] == ProductFilePath.query.all()

        sa.session.delete(path2)
        sa.session.commit()

    with app.app_context():
        product = Product.query.filter_by(name="product").one_or_none()
        assert product is not None
        assert [] == ProductFilePath.query.all()


##__________________________________________________________________||
def test_repr(app_empty):
    app = app_empty
    path = ProductFilePath(path="/abcdef/abcdef/abcdef/abcdef/abcdef/abcdef/path.txt")
    assert repr(path) == "<ProductFilePath '...f/abcdef/path.txt'>"
    path = ProductFilePath(path="path.txt")
    assert repr(path) == "<ProductFilePath 'path.txt'>"
    path = ProductFilePath()
    assert repr(path) == "<ProductFilePath None>"

##__________________________________________________________________||



import textwrap
from async_asgi_testclient import TestClient

from a2wsgi import WSGIMiddleware

import pytest
import unittest.mock as mock


##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_authenticate(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.schema.github.mutation.authenticate", y)
    yield y


##__________________________________________________________________||
@pytest.mark.asyncio
async def test_auth(app, mock_authenticate):

    query = textwrap.dedent(
        """
        mutation AuthenticateWithGitHub($code: String!) {
          authenticateWithGitHub(code: $code) {
            authPayload {
              token
            }
          }
        }
    """[
            1:
        ]
    )

    variables = {"code": "h443xg9c"}

    return_value = {"access_token": "jpdq74xt", "token_type": "bearer", "scope": ""}
    mock_authenticate.return_value = dict(return_value)

    expected = {"authenticateWithGitHub": {"authPayload": {"token": "jpdq74xt"}}}

    app = WSGIMiddleware(app)  # convert a wsgi app to an asgi app

    data = {"query": query, "variables": variables}

    headers = {"Content-Type:": "application/json"}

    async with TestClient(app) as client:
        resp = await client.post("/graphql", json=data, headers=headers)

    assert {"data": expected} == resp.json()
    assert [mock.call("h443xg9c")] == mock_authenticate.call_args_list


##__________________________________________________________________||



"""add tables github_

Revision ID: 3f6526e7e5ff
Revises: d5e539d43045
Create Date: 2020-12-17 11:24:16.122319

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '3f6526e7e5ff'
down_revision = 'd5e539d43045'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('github_accepted_orgs',
    sa.Column('org_id', sa.Integer(), nullable=False),
    sa.Column('login', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('org_id')
    )
    op.create_table('github_users',
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('login', sa.Text(), nullable=False),
    sa.PrimaryKeyConstraint('user_id')
    )
    op.create_table('github_org_memberships',
    sa.Column('entry_id', sa.Integer(), nullable=False),
    sa.Column('org_id', sa.Integer(), nullable=False),
    sa.Column('member_id', sa.Integer(), nullable=False),
    sa.ForeignKeyConstraint(['member_id'], ['github_users.user_id'], ),
    sa.ForeignKeyConstraint(['org_id'], ['github_accepted_orgs.org_id'], ),
    sa.PrimaryKeyConstraint('entry_id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('github_org_memberships')
    op.drop_table('github_users')
    op.drop_table('github_accepted_orgs')
    # ### end Alembic commands ###



from ...db.sa import sa


##__________________________________________________________________||
class GitHubUser(sa.Model):
    __tablename__ = "github_users"
    user_id = sa.Column(sa.Integer(), primary_key=True)
    git_hub_id = sa.Column(sa.Text(), unique=True, nullable=False)
    login = sa.Column(sa.Text(), unique=True, nullable=False)
    name = sa.Column(sa.Text())
    avatar_url = sa.Column(sa.Text())
    url = sa.Column(sa.Text())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.login!r}>"


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_success[one] 1'] = {
    'data': {
        'gitHubViewer': {
            'avatarUrl': 'avatar.com/user1',
            'login': 'user1',
            'name': 'User One'
        }
    }
}



import pytest

from acondbs.db.sa import sa
from acondbs.models import GitHubOrg, GitHubUser, GitHubOrgMembership


##__________________________________________________________________||
@pytest.fixture
def app(app_empty):
    y = app_empty

    org1 = GitHubOrg(login="org1", git_hub_id="012:Organization1")
    user1 = GitHubUser(login="user1", git_hub_id="04:User1")
    membership = GitHubOrgMembership(org=org1, member=user1)

    with y.app_context():
        sa.session.add(membership)
        sa.session.commit()
    yield y


##__________________________________________________________________||



from acondbs.db.sa import sa
from acondbs.models import ProductRelationType


##__________________________________________________________________||
def test_column(app_empty):
    app = app_empty

    with app.app_context():
        model = ProductRelationType(
            name="parent",
            indef_article="a",
            singular="parent",
            plural="parents",
        )
        sa.session.add(model)
        sa.session.commit()
        assert model.type_id
        type_id = model.type_id

    with app.app_context():
        model = ProductRelationType.query.filter_by(type_id=type_id).one()
        assert model.type_id == type_id
        assert model.name == "parent"
        assert model.indef_article == "a"
        assert model.singular == "parent"
        assert model.plural == "parents"


##__________________________________________________________________||



import json

from acondbs import ops
from acondbs.models import WebConfig


def test_new(app_empty):
    app = app_empty

    config_json = json.dumps(
        {
            "headTitle": "Head Title",
            "toolbarTitle": "Toolbar Title",
        },
        indent=2,
    )

    with app.app_context():
        ops.save_web_config(json=config_json)
        ops.commit()

    with app.app_context():
        model = WebConfig.query.one()
        assert model.json == config_json


def test_update(app_empty):
    app = app_empty

    config_json = json.dumps(
        {
            "headTitle": "Head Title",
            "toolbarTitle": "Toolbar Title",
        },
        indent=2,
    )

    with app.app_context():
        ops.save_web_config(json=config_json)
        ops.commit()

    config_json = json.dumps(
        {
            "headTitle": "Updated Head Title",
            "toolbarTitle": "Updated Toolbar Title",
        },
        indent=2,
    )

    with app.app_context():
        ops.save_web_config(json=config_json)
        ops.commit()

    with app.app_context():
        model = WebConfig.query.one()
        assert model.json == config_json



FRAGMENT_PRODUCT_TYPE = '''
fragment fragmentProductType on ProductType {
  typeId
  name
  order
  indefArticle
  singular
  plural
  icon
  fields {
    edges {
      node {
        type_ {
          name
        }
        field {
          name
          type_
        }
      }
    }
  }
  products {
    edges {
      node {
        name
      }
    }
  }
}
'''



import graphene
from graphene import relay

from . import (
    version as version_,
    web,
    auth,
    github,
    misc,
    product as p,
)


##__________________________________________________________________||
class QueryPublic(graphene.ObjectType):
    web_config = web.query.web_config_field
    is_signed_in = auth.query.is_signed_in_field
    git_hub_o_auth_app_info = github.query.git_hub_o_auth_app_info_field


class MutationPublic(graphene.ObjectType):
    authenticate_with_git_hub = github.mutation.AuthenticateWithGitHub.Field()
    create_log = misc.mutation.CreateLog.Field()


##__________________________________________________________________||
class QueryPrivate(QueryPublic):

    #
    version = version_.version_field
    alembic_version = version_.alembic_version_field

    node = relay.Node.Field()

    #
    git_hub_viewer = github.query.git_hub_viewer_field

    #
    is_admin = auth.query.is_admin_field

    #
    all_products = p.query.all_products_field
    all_product_file_paths = p.query.all_product_file_paths_field
    all_product_relations = p.query.all_product_relations_field
    all_product_relation_types = p.query.all_product_relation_types_field
    all_product_types = p.query.all_product_types_field
    all_fields = p.query.all_fields_field

    product = p.query.product_field
    product_relation = p.query.product_relation_field
    product_relation_type = p.query.product_relation_type_field
    product_type = p.query.product_type_field
    field = p.query.field_field


class MutationPrivate(MutationPublic):

    #
    create_product = p.mutation.CreateProduct.Field()
    delete_product = p.mutation.DeleteProduct.Field()
    update_product = p.mutation.UpdateProduct.Field()

    create_product_file_path = p.mutation.CreateProductFilePath.Field()
    delete_product_file_path = p.mutation.DeleteProductFilePath.Field()
    update_product_file_path = p.mutation.UpdateProductFilePath.Field()

    create_product_relation = p.mutation.CreateProductRelation.Field()
    delete_product_relation = p.mutation.DeleteProductRelation.Field()

    create_product_relation_types = (
        p.mutation.CreateProductRelationTypes.Field()
    )
    delete_product_relation_types = (
        p.mutation.DeleteProductRelationTypes.Field()
    )
    update_product_relation_type = p.mutation.UpdateProductRelationType.Field()

    create_product_type = p.mutation.CreateProductType.Field()
    delete_product_type = p.mutation.DeleteProductType.Field()
    update_product_type = p.mutation.UpdateProductType.Field()

    create_field = p.mutation.CreateField.Field()
    delete_field = p.mutation.DeleteField.Field()
    update_field = p.mutation.UpdateField.Field()


##__________________________________________________________________||
class QueryAdmin(QueryPrivate):

    #
    all_git_hub_orgs = github.query.all_git_hub_orgs_field
    all_git_hub_tokens = github.query.all_git_hub_tokens_field
    all_git_hub_users = github.query.all_git_hub_users_field

    #
    all_logs = misc.query.all_logs_field

    #
    log = misc.query.log_field


class MutationAdmin(MutationPrivate):

    #
    add_git_hub_org = github.mutation.AddGitHubOrg.Field()
    delete_git_hub_org = github.mutation.DeleteGitHubOrg.Field()

    add_git_hub_admin_app_token = (
        github.mutation.AddGitHubAdminAppToken.Field()
    )
    delete_git_hub_admin_app_token = (
        github.mutation.DeleteGitHubAdminAppToken.Field()
    )

    update_git_hub_org_member_lists = (
        github.mutation.UpdateGitHubOrgMemberLists.Field()
    )

    delete_log = misc.mutation.DeleteLog.Field()

    save_web_config = web.mutation.SaveWebconfig.Field()


##__________________________________________________________________||
class Query(QueryAdmin):
    pass


class Mutation(MutationAdmin):
    pass


##__________________________________________________________________||
schema_public = graphene.Schema(query=QueryPublic, mutation=MutationPublic)
schema_private = graphene.Schema(query=QueryPrivate, mutation=MutationPrivate)
schema_admin = graphene.Schema(query=QueryAdmin, mutation=MutationAdmin)

schema = graphene.Schema(query=Query, mutation=Mutation)

##__________________________________________________________________||



import graphene

from . import type_

from ... import ops


class SaveWebconfig(graphene.Mutation):
    class Arguments:
        json = graphene.String(required=True)

    ok = graphene.Boolean()
    web_config = graphene.Field(lambda: type_.WebConfig)

    def mutate(root, info, json):
        model = ops.save_web_config(json=json)
        ops.commit()
        ok = True
        return SaveWebconfig(web_config=model, ok=ok)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[dojocat] 1'] = {
    'data': {
        'isSignedIn': True
    }
}

snapshots['test_schema[no-token] 1'] = {
    'data': {
        'isSignedIn': False
    }
}

snapshots['test_schema[octocat] 1'] = {
    'data': {
        'isSignedIn': True
    }
}

snapshots['test_schema[wrong-token] 1'] = {
    'data': {
        'isSignedIn': False
    }
}



"""enter data in attribute tables

Revision ID: cceebfa64b83
Revises: e58cc402c887
Create Date: 2021-11-03 16:38:07.931890

"""
from alembic import op

from sqlalchemy.orm.session import Session

from sqlalchemy import MetaData
from flask_sqlalchemy import SQLAlchemy

# revision identifiers, used by Alembic.
revision = "cceebfa64b83"
down_revision = "e58cc402c887"
branch_labels = None
depends_on = None

#
convention = {
    "ix": "ix_%(column_0_label)s",
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    # "ck": "ck_%(table_name)s_%(constraint_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
    "pk": "pk_%(table_name)s",
}
metadata = MetaData(naming_convention=convention)

sa = SQLAlchemy(metadata=metadata)

#
from sqlalchemy.orm import declared_attr
from sqlalchemy.orm import declarative_mixin
import enum


@declarative_mixin
class AttributeBase:
    iid = sa.Column(sa.Integer(), primary_key=True)

    @declared_attr
    def product_id(self):
        return sa.Column(
            sa.Integer(),
            sa.ForeignKey("products.product_id"),
            nullable=False,
        )

    @declared_attr
    def product(self):
        return sa.relationship(
            "Product",
            backref=sa.backref(
                self.backref_column,
                cascade="all, delete-orphan",
            ),
        )

    @declared_attr
    def type_field_association_iid(self):
        return sa.Column(
            sa.Integer(),
            sa.ForeignKey("type_field_association.iid"),
            nullable=True,
        )
        # TODO: Make nullable False

    @declared_attr
    def type_field_association(self):
        return sa.relationship(
            "TypeFieldAssociation",
            backref=sa.backref(
                self.backref_column,  # TODO: need to change
                cascade="all, delete-orphan",
            ),
        )

    @declared_attr
    def field_id(self):
        return sa.Column(
            sa.Integer(),
            sa.ForeignKey("field.field_id"),
            nullable=True,
        )
        # TODO: remove, replace with type_field_association_iid

    @declared_attr
    def field(self):
        return sa.relationship(
            "Field",
            backref=sa.backref(
                self.backref_column,
                cascade="all, delete-orphan",
            ),
        )
        # TODO: remove, replace with type_field_association

    def __repr__(self):
        return (
            f"<{self.__class__.__name__} {self.field_name!r}: {self.value!r}>"
        )

    @property
    def field_name(self):
        try:
            return self.field.name
        except BaseException:
            return self.field


class AttributeUnicodeText(AttributeBase, sa.Model):
    __tablename__ = "attribute_unicode_text"
    backref_column = "attributes_unicode_text"
    value = sa.Column(sa.UnicodeText())


class AttributeBoolean(AttributeBase, sa.Model):
    __tablename__ = "attribute_boolean"
    backref_column = "attributes_boolean"
    value = sa.Column(sa.Boolean())


class AttributeInteger(AttributeBase, sa.Model):
    __tablename__ = "attribute_integer"
    backref_column = "attributes_integer"
    value = sa.Column(sa.Integer())


class AttributeFloat(AttributeBase, sa.Model):
    __tablename__ = "attribute_float"
    backref_column = "attributes_float"
    value = sa.Column(sa.Float())


class AttributeDate(AttributeBase, sa.Model):
    __tablename__ = "attribute_date"
    backref_column = "attributes_date"
    value = sa.Column(sa.Date())


class AttributeDateTime(AttributeBase, sa.Model):
    __tablename__ = "attribute_date_time"
    backref_column = "attributes_date_time"
    value = sa.Column(sa.DateTime())


class AttributeTime(AttributeBase, sa.Model):
    __tablename__ = "attribute_time"
    backref_column = "attributes_time"
    value = sa.Column(sa.Time())


class FieldType(enum.Enum):
    # SQLAlchemy generic types: https://docs.sqlalchemy.org/en/14/core/type_basics.html#generic-types
    UnicodeText = 1
    Boolean = 2
    Integer = 3
    Float = 4
    Date = 5
    DateTime = 6
    Time = 7

    @property
    def attribute_class(self):
        return FIELDTYPE_ATTRIBUTECLASS_MAP[self]


FIELDTYPE_ATTRIBUTECLASS_MAP = {
    FieldType.UnicodeText: AttributeUnicodeText,
    FieldType.Boolean: AttributeBoolean,
    FieldType.Integer: AttributeInteger,
    FieldType.Float: AttributeFloat,
    FieldType.Date: AttributeDate,
    FieldType.DateTime: AttributeDateTime,
    FieldType.Time: AttributeTime,
}


saEnumFieldType = sa.Enum(FieldType)  # to be imported in another module


class Field(sa.Model):
    __tablename__ = "field"
    field_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.UnicodeText(), nullable=False)
    type_ = sa.Column(saEnumFieldType, nullable=False)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r} {self.type_name!r}>"

    @property
    def type_name(self):
        # used in __repr__()
        try:
            return self.type_.name
        except BaseException:
            return self.type_


class TypeFieldAssociation(sa.Model):
    __tablename__ = "type_field_association"
    iid = sa.Column(sa.Integer(), primary_key=True)
    type_id = sa.Column(
        sa.Integer(),
        sa.ForeignKey("product_types.type_id"),
    )
    field_id = sa.Column(
        sa.Integer(),
        sa.ForeignKey("field.field_id"),
        nullable=False,
    )
    type_ = sa.relationship(
        "ProductType",
        backref=sa.backref("fields", cascade="all, delete-orphan"),
    )
    field = sa.relationship(
        "Field",
        backref=sa.backref("entry_types"),
    )

    __table_args__ = (
        sa.UniqueConstraint("type_id", "field_id", name="_type_field"),
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.type_name!r} {self.field_name!r}>"

    @property
    def type_name(self):
        # used in __repr__()
        try:
            return self.type_.name
        except BaseException:
            return self.type_

    @property
    def field_name(self):
        # used in __repr__()
        try:
            return self.field.name
        except BaseException:
            return self.field


class ProductType(sa.Model):
    __tablename__ = "product_types"
    type_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.Text(), nullable=False, unique=True, index=True)
    order = sa.Column(sa.Integer())
    indef_article = sa.Column(sa.Text())
    singular = sa.Column(sa.Text())
    plural = sa.Column(sa.Text())
    icon = sa.Column(sa.Text())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"


class Product(sa.Model):
    __tablename__ = "products"
    product_id = sa.Column(sa.Integer(), primary_key=True)
    type_id = sa.Column(sa.ForeignKey("product_types.type_id"), nullable=False)
    type_ = sa.relationship("ProductType", backref=sa.backref("products"))
    name = sa.Column(sa.Text(), nullable=False)
    contact = sa.Column(sa.Text())
    date_produced = sa.Column(sa.Date())
    produced_by = sa.Column(sa.Text())
    time_posted = sa.Column(
        sa.DateTime(), default=lambda: datetime.datetime.now()
    )
    posted_by = sa.Column(sa.Text())
    posting_git_hub_user_id = sa.Column(sa.ForeignKey("github_users.user_id"))
    posting_git_hub_user = sa.relationship(
        "GitHubUser",
        foreign_keys=[posting_git_hub_user_id],
        backref=sa.backref("posted_products", cascade="all"),
    )
    time_updated = sa.Column(sa.DateTime())
    updated_by = sa.Column(sa.Text())
    updating_git_hub_user_id = sa.Column(sa.ForeignKey("github_users.user_id"))
    updating_git_hub_user = sa.relationship(
        "GitHubUser",
        foreign_keys=[updating_git_hub_user_id],
        backref=sa.backref("updated_products", cascade="all"),
    )
    note = sa.Column(sa.Text())
    __table_args__ = (
        sa.UniqueConstraint("type_id", "name", name="_type_id_name"),
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"


class GitHubUser(sa.Model):
    __tablename__ = "github_users"
    user_id = sa.Column(sa.Integer(), primary_key=True)
    git_hub_id = sa.Column(sa.Text(), unique=True, nullable=False)
    login = sa.Column(sa.Text(), unique=True, nullable=False)
    name = sa.Column(sa.Text())
    avatar_url = sa.Column(sa.Text())
    url = sa.Column(sa.Text())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.login!r}>"


#
def upgrade():
    session = Session(bind=op.get_bind())

    with session.no_autoflush:
        for field_type in FieldType.__members__.values():
            for attr in session.query(field_type.attribute_class):
                assoc = session.query(TypeFieldAssociation).filter_by(
                    type_=attr.product.type_, field=attr.field
                ).one()
                attr.type_field_association = assoc
    session.commit()


def downgrade():
    pass



from ..fragments import FRAGMENT_FIELD

MUTATION_UPDATE_FIELD = (
    """
mutation UpdateField($fieldId: Int!, $input: UpdateFieldInput!) {
  updateField(fieldId: $fieldId, input: $input) {
    ok
    field {
      ...fragmentField
    }
  }
}
"""
    + FRAGMENT_FIELD
)



from ..fragments import FRAGMENT_FIELD

QUERY_FIELD = (
    """
query Field($fieldId: Int, $name: String) {
  field(fieldId: $fieldId, name: $name) {
    ...fragmentField
  }
}
"""
    + FRAGMENT_FIELD
)



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema[by-id] 1'] = {
    'data': {
        'log': {
            'id_': '1',
            'level': 'DEBUG',
            'message': 'A debug message!',
            'time': '2021-01-04T14:32:20'
        }
    }
}



from ..fragments import FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION

QUERY_ALL_PRODUCT_RELATION_TYPES = (
    """
{
  allProductRelationTypes {
    ...fragmentProductRelationTypeConnection
  }
}
"""
    + FRAGMENT_PRODUCT_RELATION_TYPE_CONNECTION
)



import pytest

import datetime

from acondbs import ops


##__________________________________________________________________||
@pytest.fixture
def app(app_users):

    y = app_users

    # Relation types:
    #   parent <-> child
    #   plaintiff <-> defendant

    # Relations:
    #   map1 -> beam1
    #     |        |
    #     +--------+---> beam2

    with y.app_context():
        ops.create_product_relation_type(
            type_={
                "type_id": 1,
                "name": "parent",
                "indef_article": "a",
                "singular": "parent",
                "plural": "parents",
            },
            reverse={
                "type_id": 2,
                "name": "child",
                "indef_article": "a",
                "singular": "child",
                "plural": "children",
            },
        )
        ops.create_product_relation_type(
            type_={
                "type_id": 3,
                "name": "plaintiff",
                "indef_article": "a",
                "singular": "plaintiff",
                "plural": "plaintiffs",
            },
            reverse={
                "type_id": 4,
                "name": "defendant",
                "indef_article": "a",
                "singular": "defendant",
                "plural": "defendants",
            },
        )

        ops.create_product_type(type_id=1, name="map")
        ops.create_product_type(type_id=2, name="beam")

        ops.commit()

    with y.app_context():
        ops.create_product(
            product_id=1,
            type_id=1,
            name="map1",
            date_produced=datetime.date(2020, 2, 1),
        )
        ops.create_product(
            product_id=4,
            type_id=2,
            name="beam1",
            date_produced=datetime.date(2020, 2, 5),
        )
        ops.create_product(
            product_id=5,
            type_id=2,
            name="beam2",
            date_produced=datetime.date(2020, 3, 4),
        )
        ops.commit()

    with y.app_context():
        ops.create_product_relation(
            type_id=1,
            self_product_id=4,
            other_product_id=1,
        )
        ops.create_product_relation(
            type_id=1,
            self_product_id=5,
            other_product_id=1,
        )
        ops.create_product_relation(
            type_id=1,
            self_product_id=5,
            other_product_id=4,
        )
        ops.commit()

    yield y


##__________________________________________________________________||




import pytest
import unittest.mock as mock

from acondbs.github import query

##__________________________________________________________________||
@pytest.fixture(autouse=True)
def mock_requests(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.github.call.requests", y)
    yield y


##__________________________________________________________________||
def test_success(mock_requests):

    token = '4d5dc8b74eccdf65859d6ac64358a3a98300c351'
    # random string generated with
    # ''.join(random.choice('0123456789abcdef') for i in range(40))

    mock_requests.post().json.return_value = {
        'data': {
            'viewer': {
                "id": "MDQ6VXNlcjU4MzIzMQ==",
                "login": "octocat",
                "name": "The Octocat",
                "avatarUrl": "https://avatars3.githubusercontent.com/u/583231?u=a59fef2a493e2b67dd13754231daf220c82ba84d&v=4",
                "url": "https://github.com/octocat"
            }
        }
    }

    expected = {
        "id": "04:User583231",
        "login": "octocat",
        "name": "The Octocat",
        "avatarUrl": "https://avatars3.githubusercontent.com/u/583231?u=a59fef2a493e2b67dd13754231daf220c82ba84d&v=4",
        "url": "https://github.com/octocat"
    }

    actual = query.viewer(token)

    assert expected == actual

##__________________________________________________________________||
def test_bad_credentials(mock_requests):

    token = '4d5dc8b74eccdf65859d6ac64358a3a98300c351'
    # random string generated with
    # ''.join(random.choice('0123456789abcdef') for i in range(40))

    mock_requests.post().json.return_value = {
        'message': 'Bad credentials',
        'documentation_url': 'https://docs.github.com/graphql'
    }

    with pytest.raises(Exception) as e:
        query.viewer(token)

    expected = {
        'message': 'Bad credentials',
        'documentation_url': 'https://docs.github.com/graphql'
    }

    assert expected == e.value.args[0]

##__________________________________________________________________||



import datetime
import pytest

# from sqlalchemy import exc

from acondbs import ops
from acondbs.ops.product import _normalize_paths
from acondbs.models import Product, FieldType, GitHubUser


##__________________________________________________________________||
# fmt: off
params = [
    pytest.param([], [], id="empty"),
    pytest.param(["/a/b/c"], ["/a/b/c"], id="one"),
    pytest.param([" /a/b/c  "], ["/a/b/c"], id="unstripped"),
    pytest.param(["/d/e", "/a/b/c", "/f/g"], ["/d/e", "/a/b/c", "/f/g"], id="multiple"),
    pytest.param(["/a/b/c", "", "/f/g"], ["/a/b/c", "/f/g"], id="empty-member"),
    pytest.param(["/a/b/c", "  ", "/f/g"], ["/a/b/c", "/f/g"], id="empty-unstripped-member"),
    pytest.param(["/d/e", "/a/b/c", "/d/e"], ["/d/e", "/a/b/c"], id="duplicate"),
    pytest.param(["  /d/e ", "/a/b/c", "/d/e"], ["/d/e", "/a/b/c"], id="duplicate-unstripped"),
]
# fmt: on


@pytest.mark.parametrize("paths, ret", params)
def test_normalize_paths(paths, ret):
    assert _normalize_paths(paths) == ret


##__________________________________________________________________||
params = [
    pytest.param(None, id="none"),
    pytest.param({}, id="empty"),
    pytest.param({1: "An attribute"}, id="one"),
    pytest.param({1: "An attribute", 3: False}, id="two"),
    pytest.param(
        {
            1: "An attribute",
            3: False,
            4: True,
            5: 512,
            6: 2.34,
            8: datetime.datetime(2021, 10, 15, 16, 55, 21),
        },
        id="all",
    ),
]


@pytest.mark.parametrize("attributes", params)
def test_create_attributes(app, attributes):
    return _test_create(app, attributes=attributes)


params = [
    pytest.param(None, id="none"),
    pytest.param([], id="empty"),
    pytest.param([{"type_id": 1, "product_id": 2}], id="one"),
    pytest.param(
        [{"type_id": 2, "product_id": 1}, {"type_id": 1, "product_id": 3}],
        id="multiple",
    ),
]


@pytest.mark.parametrize("relations", params)
def test_create_relations(app, relations):
    return _test_create(app, relations=relations)


params = [
    pytest.param(None, id="none"),
    pytest.param([], id="empty"),
    pytest.param(["/a/b/c"], id="one"),
    pytest.param(["/d/e", "/a/b/c", "/f/g"], id="multiple"),
    pytest.param(
        ["  /d/e ", " ", "/a/b/c", "/f/g", "/d/e"], id="not-normalized"
    ),
]


@pytest.mark.parametrize("paths", params)
def test_create_paths(app, paths):
    return _test_create(app, paths=paths)


@pytest.mark.parametrize("posting_git_hub_user_id", [None, 1])
def test_create_user(app, posting_git_hub_user_id):
    return _test_create(app, posting_git_hub_user_id=posting_git_hub_user_id)


def _test_create(
    app,
    paths=None,
    relations=None,
    attributes=None,
    posting_git_hub_user_id=None,
):

    kwargs = {"type_id": 1, "name": "new-product"}

    if paths is not None:
        kwargs["paths"] = paths

    if relations is not None:
        kwargs["relations"] = relations

    if attributes is not None:
        kwargs["attributes"] = attributes

    if posting_git_hub_user_id:
        kwargs["posting_git_hub_user_id"] = posting_git_hub_user_id

    with app.app_context():
        count = Product.query.count()

        model = ops.create_product(**kwargs)
        assert model.name == "new-product"
        ops.commit()
        product_id = model.product_id
        assert product_id

    with app.app_context():
        assert Product.query.count() == (count + 1)

        model = Product.query.filter_by(product_id=product_id).one()
        assert model.name == "new-product"

        if posting_git_hub_user_id:
            posting_git_hub_user = GitHubUser.query.filter_by(
                user_id=posting_git_hub_user_id
            ).one()
            assert model.posting_git_hub_user == posting_git_hub_user
        else:
            assert model.posting_git_hub_user is None

        if paths is not None:
            expected = _normalize_paths(paths)
            actual = [p.path for p in model.paths]
            assert actual == expected
        else:
            assert model.paths == []

        if relations is not None:
            expected = [(r["type_id"], r["product_id"]) for r in relations]
            actual = [(r.type_id, r.other_product_id) for r in model.relations]
            assert actual == expected
        else:
            assert model.relations == []

        if attributes is None:
            attributes = {}
        expected_ids = [(a.iid, a.field_id) for a in model.type_.fields]
        # list of (type_field_association.iid, field.field_id)
        expected = {fid: (aid, attributes.get(fid)) for aid, fid in expected_ids}
        actual = _extract_attributes(model)
        actual_field_ids = list(actual.keys())
        assert actual == expected


##__________________________________________________________________||
def _extract_attributes(model):
    attr_names = [
        a.attribute_class.backref_column
        for a in FieldType.__members__.values()
    ]
    # e.g., 'attributes_unicode_text', 'attributes_boolean'

    attrs = [e for attr in attr_names for e in getattr(model, attr)]
    # e.g., AttributeUnicodeText

    ret = {a.field_id: (a.type_field_association_iid, a.value) for a in attrs}

    return ret


##__________________________________________________________________||
params = [
    pytest.param(None, id="none"),
    pytest.param({}, id="empty"),
    pytest.param({1: "An attribute"}, id="one"),
    pytest.param({1: "An attribute", 3: False}, id="two"),
    pytest.param(
        {
            1: "An attribute",
            3: False,
            4: True,
            5: 512,
            6: 2.34,
            8: datetime.datetime(2021, 10, 15, 16, 55, 21),
        },
        id="all",
    ),
]


@pytest.mark.parametrize("attributes", params)
def test_update_attributes(app, attributes):
    return _test_update(app, attributes=attributes)


params = [
    pytest.param(None, id="none"),
    pytest.param([], id="empty"),
    pytest.param(
        [
            {"type_id": 2, "product_id": 3},
            {"type_id": 1, "product_id": 5},
        ],
        id="same",
    ),
    pytest.param(
        [
            {"type_id": 1, "product_id": 5},
        ],
        id="remove",
    ),
    pytest.param(
        [
            {"type_id": 2, "product_id": 3},
            {"type_id": 1, "product_id": 5},
            {"type_id": 1, "product_id": 4},
        ],
        id="add",
    ),
    pytest.param(
        [
            {"type_id": 1, "product_id": 4},
            {"type_id": 2, "product_id": 3},
        ],
        id="add-remove",
    ),
    pytest.param(
        [
            {"type_id": 1, "product_id": 4},
            {"type_id": 2, "product_id": 6},
        ],
        id="replace",
    ),
]


@pytest.mark.parametrize("relations", params)
def test_update_relations(app, relations):
    return _test_update(app, relations=relations)


params = [
    pytest.param(None, id="none"),
    pytest.param([], id="empty"),
    pytest.param(["/d/e", "/a/b/c"], id="same"),
    pytest.param(["/a/b/c", "/d/e"], id="same-perm"),
    pytest.param(["/a/b/c"], id="remove"),
    pytest.param(["/d/e", "/a/b/c", "/f/g"], id="add"),
    pytest.param(["/d/e", "/f/g", "/a/b/c"], id="add-perm"),
    pytest.param(
        ["  /d/e ", " ", "/a/b/c", "/f/g", "/d/e"], id="not-normalized"
    ),
]


@pytest.mark.parametrize("paths", params)
def test_update_paths(app, paths):
    return _test_update(app, paths=paths)


@pytest.mark.parametrize("updating_git_hub_user_id", [None, 2])
def test_update_user(app, updating_git_hub_user_id):
    return _test_update(app, updating_git_hub_user_id=updating_git_hub_user_id)


def _test_update(
    app,
    paths=None,
    relations=None,
    attributes=None,
    updating_git_hub_user_id=None,
):

    product_id = 1
    kwargs = {"product_id": product_id, "name": "new-name"}

    if paths is not None:
        kwargs["paths"] = paths

    if relations is not None:
        kwargs["relations"] = relations

    if attributes is not None:
        kwargs["attributes"] = attributes

    if updating_git_hub_user_id:
        kwargs["updating_git_hub_user_id"] = updating_git_hub_user_id

    with app.app_context():
        count = Product.query.count()

        model = Product.query.filter_by(product_id=product_id).one()
        paths_old = [p.path for p in model.paths]
        path_ids_old = {p.path: p.path_id for p in model.paths}

        relations_old = {
            (r.type_id, r.other_product_id) for r in model.relations
        }
        relation_ids_old = {
            (r.type_id, r.other_product_id): r.relation_id
            for r in model.relations
        }

        attributes_old = _extract_attributes(model)

        model = ops.update_product(**kwargs)
        assert model.name == "new-name"
        ops.commit()

    with app.app_context():
        assert Product.query.count() == count

        model = Product.query.filter_by(product_id=product_id).one()
        assert model.name == "new-name"

        if updating_git_hub_user_id:
            updating_git_hub_user = GitHubUser.query.filter_by(
                user_id=updating_git_hub_user_id
            ).one()
            assert model.updating_git_hub_user == updating_git_hub_user
        else:
            assert model.updating_git_hub_user is None

        # NOTE: The order of the paths is being tested. However, it is
        # probably not possible to control the order. The order in a
        # Python list is not preserved once committed in the DB.
        if paths is not None:
            kept = [p for p in paths_old if p in paths]
            expected = kept + _normalize_paths(paths)
            expected = list(dict.fromkeys(expected))  # uniq order preserved
        else:
            expected = paths_old
        actual = [p.path for p in model.paths]
        assert actual == expected
        expected_path_ids = {
            k: v for k, v in path_ids_old.items() if k in expected
        }
        path_ids = {
            p.path: p.path_id for p in model.paths if p.path in path_ids_old
        }
        assert path_ids == expected_path_ids

        if relations is not None:
            expected = {(r["type_id"], r["product_id"]) for r in relations}
        else:
            expected = relations_old
        actual = {(r.type_id, r.other_product_id) for r in model.relations}
        assert actual == expected
        expected_relation_ids = {
            k: v for k, v in relation_ids_old.items() if k in expected
        }
        relation_ids = {
            (r.type_id, r.other_product_id): r.relation_id
            for r in model.relations
            if (r.type_id, r.other_product_id) in relation_ids_old
        }
        assert relation_ids == expected_relation_ids

        if attributes is None:
            attributes = {}
        expected = attributes_old.copy()
        update = {k: (expected[k][0], v) for k, v in attributes.items()}
        expected.update(update)
        actual = _extract_attributes(model)
        assert actual == expected


##__________________________________________________________________||
def test_delete(app):

    with app.app_context():
        model = ops.create_product(type_id=1, name="to_be_deleted")
        ops.commit()
        product_id = model.product_id

    with app.app_context():
        count = Product.query.count()
        model = ops.delete_product(product_id=product_id)
        ops.commit()

    with app.app_context():
        model = Product.query.filter_by(product_id=product_id).one_or_none()
        assert model is None
        assert Product.query.count() == (count - 1)


##__________________________________________________________________||



QUERY_ALL_PRODUCT_FILE_PATHS = """
{
  allProductFilePaths {
    totalCount
    edges {
      node {
        pathId
        path
        note
        product {
          name
        }
      }
    }
  }
}
"""



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_selection[False-False] 1'] = {
    'data': {
        '__schema': {
            'mutationType': {
                'fields': [
                    {
                        'name': 'authenticateWithGitHub'
                    },
                    {
                        'name': 'createLog'
                    }
                ]
            },
            'queryType': {
                'fields': [
                    {
                        'name': 'webConfig'
                    },
                    {
                        'name': 'isSignedIn'
                    },
                    {
                        'name': 'gitHubOAuthAppInfo'
                    }
                ]
            },
            'subscriptionType': None
        }
    }
}

snapshots['test_schema_selection[False-True] 1'] = {
    'data': {
        '__schema': {
            'mutationType': {
                'fields': [
                    {
                        'name': 'authenticateWithGitHub'
                    },
                    {
                        'name': 'createLog'
                    },
                    {
                        'name': 'createProduct'
                    },
                    {
                        'name': 'deleteProduct'
                    },
                    {
                        'name': 'updateProduct'
                    },
                    {
                        'name': 'createProductFilePath'
                    },
                    {
                        'name': 'deleteProductFilePath'
                    },
                    {
                        'name': 'updateProductFilePath'
                    },
                    {
                        'name': 'createProductRelation'
                    },
                    {
                        'name': 'deleteProductRelation'
                    },
                    {
                        'name': 'createProductRelationTypes'
                    },
                    {
                        'name': 'deleteProductRelationTypes'
                    },
                    {
                        'name': 'updateProductRelationType'
                    },
                    {
                        'name': 'createProductType'
                    },
                    {
                        'name': 'deleteProductType'
                    },
                    {
                        'name': 'updateProductType'
                    },
                    {
                        'name': 'createField'
                    },
                    {
                        'name': 'deleteField'
                    },
                    {
                        'name': 'updateField'
                    }
                ]
            },
            'queryType': {
                'fields': [
                    {
                        'name': 'webConfig'
                    },
                    {
                        'name': 'isSignedIn'
                    },
                    {
                        'name': 'gitHubOAuthAppInfo'
                    },
                    {
                        'name': 'version'
                    },
                    {
                        'name': 'alembicVersion'
                    },
                    {
                        'name': 'isAdmin'
                    },
                    {
                        'name': 'gitHubViewer'
                    },
                    {
                        'name': 'allProducts'
                    },
                    {
                        'name': 'allProductTypes'
                    },
                    {
                        'name': 'allProductRelations'
                    },
                    {
                        'name': 'allProductRelationTypes'
                    },
                    {
                        'name': 'allProductFilePaths'
                    },
                    {
                        'name': 'allFields'
                    },
                    {
                        'name': 'product'
                    },
                    {
                        'name': 'productType'
                    },
                    {
                        'name': 'productRelation'
                    },
                    {
                        'name': 'productRelationType'
                    },
                    {
                        'name': 'field'
                    },
                    {
                        'name': 'node'
                    }
                ]
            },
            'subscriptionType': None
        }
    }
}

snapshots['test_schema_selection[True-False] 1'] = {
    'data': {
        '__schema': {
            'mutationType': {
                'fields': [
                    {
                        'name': 'authenticateWithGitHub'
                    },
                    {
                        'name': 'createLog'
                    },
                    {
                        'name': 'createProduct'
                    },
                    {
                        'name': 'deleteProduct'
                    },
                    {
                        'name': 'updateProduct'
                    },
                    {
                        'name': 'createProductFilePath'
                    },
                    {
                        'name': 'deleteProductFilePath'
                    },
                    {
                        'name': 'updateProductFilePath'
                    },
                    {
                        'name': 'createProductRelation'
                    },
                    {
                        'name': 'deleteProductRelation'
                    },
                    {
                        'name': 'createProductRelationTypes'
                    },
                    {
                        'name': 'deleteProductRelationTypes'
                    },
                    {
                        'name': 'updateProductRelationType'
                    },
                    {
                        'name': 'createProductType'
                    },
                    {
                        'name': 'deleteProductType'
                    },
                    {
                        'name': 'updateProductType'
                    },
                    {
                        'name': 'createField'
                    },
                    {
                        'name': 'deleteField'
                    },
                    {
                        'name': 'updateField'
                    },
                    {
                        'name': 'addGitHubOrg'
                    },
                    {
                        'name': 'deleteGitHubOrg'
                    },
                    {
                        'name': 'addGitHubAdminAppToken'
                    },
                    {
                        'name': 'deleteGitHubAdminAppToken'
                    },
                    {
                        'name': 'updateGitHubOrgMemberLists'
                    },
                    {
                        'name': 'deleteLog'
                    },
                    {
                        'name': 'saveWebConfig'
                    }
                ]
            },
            'queryType': {
                'fields': [
                    {
                        'name': 'webConfig'
                    },
                    {
                        'name': 'isSignedIn'
                    },
                    {
                        'name': 'gitHubOAuthAppInfo'
                    },
                    {
                        'name': 'version'
                    },
                    {
                        'name': 'alembicVersion'
                    },
                    {
                        'name': 'isAdmin'
                    },
                    {
                        'name': 'gitHubViewer'
                    },
                    {
                        'name': 'allProducts'
                    },
                    {
                        'name': 'allProductTypes'
                    },
                    {
                        'name': 'allProductRelations'
                    },
                    {
                        'name': 'allProductRelationTypes'
                    },
                    {
                        'name': 'allProductFilePaths'
                    },
                    {
                        'name': 'allFields'
                    },
                    {
                        'name': 'product'
                    },
                    {
                        'name': 'productType'
                    },
                    {
                        'name': 'productRelation'
                    },
                    {
                        'name': 'productRelationType'
                    },
                    {
                        'name': 'field'
                    },
                    {
                        'name': 'node'
                    },
                    {
                        'name': 'allGitHubOrgs'
                    },
                    {
                        'name': 'allGitHubUsers'
                    },
                    {
                        'name': 'allGitHubTokens'
                    },
                    {
                        'name': 'allLogs'
                    },
                    {
                        'name': 'log'
                    }
                ]
            },
            'subscriptionType': None
        }
    }
}

snapshots['test_schema_selection[True-True] 1'] = {
    'data': {
        '__schema': {
            'mutationType': {
                'fields': [
                    {
                        'name': 'authenticateWithGitHub'
                    },
                    {
                        'name': 'createLog'
                    },
                    {
                        'name': 'createProduct'
                    },
                    {
                        'name': 'deleteProduct'
                    },
                    {
                        'name': 'updateProduct'
                    },
                    {
                        'name': 'createProductFilePath'
                    },
                    {
                        'name': 'deleteProductFilePath'
                    },
                    {
                        'name': 'updateProductFilePath'
                    },
                    {
                        'name': 'createProductRelation'
                    },
                    {
                        'name': 'deleteProductRelation'
                    },
                    {
                        'name': 'createProductRelationTypes'
                    },
                    {
                        'name': 'deleteProductRelationTypes'
                    },
                    {
                        'name': 'updateProductRelationType'
                    },
                    {
                        'name': 'createProductType'
                    },
                    {
                        'name': 'deleteProductType'
                    },
                    {
                        'name': 'updateProductType'
                    },
                    {
                        'name': 'createField'
                    },
                    {
                        'name': 'deleteField'
                    },
                    {
                        'name': 'updateField'
                    },
                    {
                        'name': 'addGitHubOrg'
                    },
                    {
                        'name': 'deleteGitHubOrg'
                    },
                    {
                        'name': 'addGitHubAdminAppToken'
                    },
                    {
                        'name': 'deleteGitHubAdminAppToken'
                    },
                    {
                        'name': 'updateGitHubOrgMemberLists'
                    },
                    {
                        'name': 'deleteLog'
                    },
                    {
                        'name': 'saveWebConfig'
                    }
                ]
            },
            'queryType': {
                'fields': [
                    {
                        'name': 'webConfig'
                    },
                    {
                        'name': 'isSignedIn'
                    },
                    {
                        'name': 'gitHubOAuthAppInfo'
                    },
                    {
                        'name': 'version'
                    },
                    {
                        'name': 'alembicVersion'
                    },
                    {
                        'name': 'isAdmin'
                    },
                    {
                        'name': 'gitHubViewer'
                    },
                    {
                        'name': 'allProducts'
                    },
                    {
                        'name': 'allProductTypes'
                    },
                    {
                        'name': 'allProductRelations'
                    },
                    {
                        'name': 'allProductRelationTypes'
                    },
                    {
                        'name': 'allProductFilePaths'
                    },
                    {
                        'name': 'allFields'
                    },
                    {
                        'name': 'product'
                    },
                    {
                        'name': 'productType'
                    },
                    {
                        'name': 'productRelation'
                    },
                    {
                        'name': 'productRelationType'
                    },
                    {
                        'name': 'field'
                    },
                    {
                        'name': 'node'
                    },
                    {
                        'name': 'allGitHubOrgs'
                    },
                    {
                        'name': 'allGitHubUsers'
                    },
                    {
                        'name': 'allGitHubTokens'
                    },
                    {
                        'name': 'allLogs'
                    },
                    {
                        'name': 'log'
                    }
                ]
            },
            'subscriptionType': None
        }
    }
}



from ...db.sa import sa


##__________________________________________________________________||
class TypeFieldAssociation(sa.Model):
    """Many-to-many relation between ProductType and Field


    This class defines which fields each product type has.

    When a product type is deleted, all related instances of this
    class will be automatically deleted ("delete-orphan").


    On the other hand, a field is not allowed to be deleted if it is
    related to an instances of this class (as field_id is not
    nullable while "delete-orphan" is not used on field).

    """

    __tablename__ = "type_field_association"
    iid = sa.Column(sa.Integer(), primary_key=True)
    type_id = sa.Column(
        sa.Integer(),
        sa.ForeignKey("product_types.type_id"),
        nullable=False,
    )
    field_id = sa.Column(
        sa.Integer(),
        sa.ForeignKey("field.field_id"),
        nullable=False,
    )
    type_ = sa.relationship(
        "ProductType",
        backref=sa.backref("fields", cascade="all, delete-orphan"),
    )
    field = sa.relationship(
        "Field",
        backref=sa.backref("entry_types"),
    )

    __table_args__ = (
        sa.UniqueConstraint("type_id", "field_id", name="_type_field"),
    )

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.type_name!r} {self.field_name!r}>"

    @property
    def type_name(self):
        # used in __repr__()
        try:
            return self.type_.name
        except BaseException:
            return self.type_

    @property
    def field_name(self):
        # used in __repr__()
        try:
            return self.field.name
        except BaseException:
            return self.field


##__________________________________________________________________||



def shorten(text, width, placeholder="...", end=False):
    """Truncate text

    used for repr() of models
    """

    width = max(width, len(placeholder))
    if len(text) <= width:
        return text
    pos = width - len(placeholder)
    if pos == 0:
        return placeholder
    if end:
        return placeholder + text[-pos:]
    return text[:pos] + placeholder



FRAGMENT_PRODUCT_RELATION = '''
fragment fragmentProductRelation on ProductRelation {
  relationId
  type_ {
    typeId
    name
  }
  self_ {
    productId
    name
  }
  other {
    productId
    name
  }
  reverse {
    relationId
  }
}
'''



from ...db.sa import sa


##__________________________________________________________________||
class ProductType(sa.Model):
    __tablename__ = "product_types"
    type_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.Text(), nullable=False, unique=True, index=True)
    order = sa.Column(sa.Integer())
    indef_article = sa.Column(sa.Text())
    singular = sa.Column(sa.Text())
    plural = sa.Column(sa.Text())
    icon = sa.Column(sa.Text())

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r}>"


##__________________________________________________________________||



from pathlib import Path
import pytest

from acondbs import create_app
from acondbs.db.ops import (
    define_tables,
    import_tables_from_csv_files,
    export_db_to_dict_of_dict_list,
    export_db_to_csv_files,
)

from ...constants import SAMPLE_DIR


##__________________________________________________________________||
@pytest.fixture
def app():
    config_path = Path(SAMPLE_DIR, "config.py")
    database_uri = "sqlite:///:memory:"
    app = create_app(
        config_path=config_path, SQLALCHEMY_DATABASE_URI=database_uri
    )
    with app.app_context():
        define_tables()
    yield app


##__________________________________________________________________||
def test_non_empty(app, snapshot):
    csvdir = Path(SAMPLE_DIR, "csv")
    with app.app_context():
        import_tables_from_csv_files(csvdir)
    with app.app_context():
        snapshot.assert_match(export_db_to_dict_of_dict_list())


##__________________________________________________________________||
def test_empty(app, tmpdir_factory):
    csvdir = str(tmpdir_factory.mktemp("csv"))

    with app.app_context():
        export_db_to_csv_files(csvdir)

    with app.app_context():
        import_tables_from_csv_files(csvdir)


##__________________________________________________________________||



from ..models import ProductFilePath

from ..db.sa import sa

def create_product_file_path(**kwargs):
    model = ProductFilePath(**kwargs)
    sa.session.add(model)
    return model

def update_product_file_path(path_id, **kwargs):
    model = ProductFilePath.query.filter_by(path_id=path_id).one()
    for k, v in kwargs.items():
        setattr(model, k, v)
    return model

def delete_product_file_path(path_id):
    model = ProductFilePath.query.filter_by(path_id=path_id).one()
    sa.session.delete(model)
    return



"""remove column from attribute tables

Revision ID: 6bd27755b1a8
Revises: e69be63fc3b8
Create Date: 2021-11-02 13:40:15.965707

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '6bd27755b1a8'
down_revision = 'e69be63fc3b8'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('attribute_boolean', schema=None) as batch_op:
        batch_op.drop_column('name')

    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.drop_column('name')

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.drop_column('name')

    with op.batch_alter_table('attribute_float', schema=None) as batch_op:
        batch_op.drop_column('name')

    with op.batch_alter_table('attribute_integer', schema=None) as batch_op:
        batch_op.drop_column('name')

    with op.batch_alter_table('attribute_time', schema=None) as batch_op:
        batch_op.drop_column('name')

    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.drop_column('name')

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.add_column(sa.Column('name', sa.TEXT(), nullable=True))

    with op.batch_alter_table('attribute_time', schema=None) as batch_op:
        batch_op.add_column(sa.Column('name', sa.TEXT(), nullable=True))

    with op.batch_alter_table('attribute_integer', schema=None) as batch_op:
        batch_op.add_column(sa.Column('name', sa.TEXT(), nullable=True))

    with op.batch_alter_table('attribute_float', schema=None) as batch_op:
        batch_op.add_column(sa.Column('name', sa.TEXT(), nullable=True))

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.add_column(sa.Column('name', sa.TEXT(), nullable=True))

    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.add_column(sa.Column('name', sa.TEXT(), nullable=True))

    with op.batch_alter_table('attribute_boolean', schema=None) as batch_op:
        batch_op.add_column(sa.Column('name', sa.TEXT(), nullable=True))

    # ### end Alembic commands ###



from ..fragments import FRAGMENT_PRODUCT_TYPE

MUTATION_CREATE_PRODUCT_TYPE = '''
mutation CreateProductType($input: CreateProductTypeInput!) {
  createProductType(input: $input) {
    ok
    productType {
      ...fragmentProductType
    }
  }
}
''' + FRAGMENT_PRODUCT_TYPE



import enum

from ...db.sa import sa

from . import attribute


##__________________________________________________________________||
class FieldType(enum.Enum):
    # SQLAlchemy generic types: https://docs.sqlalchemy.org/en/14/core/type_basics.html#generic-types
    UnicodeText = 1
    Boolean = 2
    Integer = 3
    Float = 4
    Date = 5
    DateTime = 6
    Time = 7

    @property
    def attribute_class(self):
        return FIELDTYPE_ATTRIBUTECLASS_MAP[self]


FIELDTYPE_ATTRIBUTECLASS_MAP = {
    FieldType.UnicodeText: attribute.AttributeUnicodeText,
    FieldType.Boolean: attribute.AttributeBoolean,
    FieldType.Integer: attribute.AttributeInteger,
    FieldType.Float: attribute.AttributeFloat,
    FieldType.Date: attribute.AttributeDate,
    FieldType.DateTime: attribute.AttributeDateTime,
    FieldType.Time: attribute.AttributeTime,
}


saEnumFieldType = sa.Enum(FieldType)  # to be imported in another module


class Field(sa.Model):
    __tablename__ = "field"
    field_id = sa.Column(sa.Integer(), primary_key=True)
    name = sa.Column(sa.UnicodeText(), nullable=False)
    type_ = sa.Column(saEnumFieldType, nullable=False)

    def __repr__(self):
        return f"<{self.__class__.__name__} {self.name!r} {self.type_name!r}>"

    @property
    def type_name(self):
        # used in __repr__()
        try:
            return self.type_.name
        except BaseException:
            return self.type_


##__________________________________________________________________||



from ..db.sa import sa
from ..models import WebConfig


def save_web_config(**kwargs):
    model = WebConfig.query.one_or_none()
    if model:
        for k, v in kwargs.items():
            setattr(model, k, v)
    else:
        model = WebConfig(**kwargs)
        sa.session.add(model)
    return model



from ..models import Field, FieldType
from ..db.sa import sa


##__________________________________________________________________||
def create_field(name, type_, field_id=None):
    """Instantiate SQLAlchemy ORM model "Field"

    Parameters
    ----------
    name : str
        The name of the new field
    type_ : enum or int
        The type of the new field, e.g., unicode text, int
    field_id : int, optional
        The ID of the new field

    Returns
    -------
    object
        The created model

    """
    type_ = FieldType(type_)  # in case given by int
    model = Field(name=name, type_=type_, field_id=field_id)
    sa.session.add(model)
    return model


def update_field(field_id, name):
    """Update a field

    Only name can be changed.

    Parameters
    ----------
    field_id : int
        The ID of the field to be updated
    name : str
        The new name

    Returns
    -------
    object
        The model

    """
    model = Field.query.filter_by(field_id=field_id).one()
    model.name = name
    return model


def delete_field(field_id):
    """Delete a field

    Parameters
    ----------
    field_id : int
        The ID of the field to be deleted

    Returns
    -------
    None

    """
    model = Field.query.filter_by(field_id=field_id).one()
    sa.session.delete(model)


##__________________________________________________________________||



import pytest

import datetime

from acondbs import ops


##__________________________________________________________________||
@pytest.fixture
def app(app_users):

    y = app_users

    # map1 -> beam1
    #            |
    #            +---> beam2
    #
    # map2
    # map3

    with y.app_context():
        ops.create_product_relation_type(
            type_={"type_id": 1, "name": "parent"},
            reverse={"type_id": 2, "name": "child"},
        )

        ops.create_product_type(type_id=1, name="map")
        ops.create_product_type(type_id=2, name="beam")

        ops.commit()

    with y.app_context():
        ops.create_product(
            product_id=1,
            type_id=1,
            name="map1",
            date_produced=datetime.date(2020, 2, 1),
        )
        ops.create_product(
            product_id=2,
            type_id=1,
            name="map2",
            date_produced=datetime.date(2020, 2, 10),
        )
        ops.create_product(
            product_id=3,
            type_id=1,
            name="map3",
            date_produced=datetime.date(2020, 3, 19),
        )
        ops.create_product(
            product_id=4,
            type_id=2,
            name="beam1",
            date_produced=datetime.date(2020, 2, 5),
        )
        ops.create_product(
            product_id=5,
            type_id=2,
            name="beam2",
            date_produced=datetime.date(2020, 3, 4),
        )
        ops.commit()

    with y.app_context():
        ops.create_product_relation(
            relation_id=2,
            type_id=1,
            self_product_id=4,
            other_product_id=1,
        )
        ops.create_product_relation(
            relation_id=4,
            type_id=1,
            self_product_id=5,
            other_product_id=4,
        )
        ops.commit()

    yield y


##__________________________________________________________________||



import warnings

import pytest
import unittest.mock as mock

from acondbs.db.backup import backup_db

##__________________________________________________________________||
@pytest.fixture()
def mock_backup_db_to_github(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.db.backup.backup_db_to_github", y)
    yield y

@pytest.fixture()
def mock_backup_db_as_csv_to_github(monkeypatch):
    y = mock.Mock()
    monkeypatch.setattr("acondbs.db.backup.backup_db_as_csv_to_github", y)
    yield y

##__________________________________________________________________||
def test_backup_db(mock_backup_db_to_github, mock_backup_db_as_csv_to_github):
    backup_db()
    mock_backup_db_to_github.assert_called_once()
    mock_backup_db_as_csv_to_github.assert_called_once()

def test_exceptions(mock_backup_db_to_github, mock_backup_db_as_csv_to_github):
    mock_backup_db_to_github.side_effect = Exception()
    mock_backup_db_as_csv_to_github.side_effect = Exception()
    with warnings.catch_warnings(record=True) as w:
        backup_db()
    assert len(w) == 2
    mock_backup_db_to_github.assert_called_once()
    mock_backup_db_as_csv_to_github.assert_called_once()

##__________________________________________________________________||



from ..db.sa import sa
from ..models import ProductType, Field, TypeFieldAssociation


##__________________________________________________________________||
def create_product_type(field_ids=None, **kwargs):
    """Create a product type"""
    model = ProductType(**kwargs)
    if field_ids:
        with sa.session.no_autoflush:
            model.fields = _create_fields(field_ids)
    sa.session.add(model)
    return model


def _create_fields(field_ids):
    ids = sorted(set(field_ids))
    fields_ = [Field.query.filter_by(field_id=i).one() for i in ids]
    return [TypeFieldAssociation(field=f) for f in fields_]


##__________________________________________________________________||
def update_product_type(type_id, field_ids=None, **kwargs):
    """Update a product type"""

    model = ProductType.query.filter_by(type_id=type_id).one()

    for k, v in kwargs.items():
        setattr(model, k, v)

    if field_ids is not None:
        with sa.session.no_autoflush:
            model.fields = _update_fields(model.fields, field_ids)

    return model


def _update_fields(old_fields: list, new_field_ids: list) -> list:

    new_ids = set(new_field_ids)
    old_ids = {f.field_id for f in old_fields}

    added_ids = new_ids - old_ids

    field_dict = {f.field_id: f for f in old_fields}
    #  {field_id: TypeFieldAssociation}

    for id_ in added_ids:
        field_ = Field.query.filter_by(field_id=id_).one()
        field = TypeFieldAssociation(field=field_)
        field_dict[id_] = field

    # It is unnecessary to explicitly delete removed instances of
    # TypeFieldAssociation from the session by calling
    # sa.session.delete(). They will be automatically deleted because
    # of the cascade option of the relationship.

    return [field_dict[i] for i in sorted(new_ids)]


##__________________________________________________________________||
def delete_product_type(type_id):
    """Delete a product type"""
    model = ProductType.query.filter_by(type_id=type_id).one()
    sa.session.delete(model)


##__________________________________________________________________||



import pytest

from ...funcs import assert_query

GITHUB_TOKEN_FRAGMENT = """
fragment GitHubTokenFragment on GitHubToken {
  tokenId
  tokenMasked
  scope
  timeCreated
  user {
    login
  }
}
"""

ALL_GITHUB_TOKENS = (
    """
{
  allGitHubTokens {
    totalCount
    edges {
      node {
        ...GitHubTokenFragment
      }
    }
  }
}
"""
    + GITHUB_TOKEN_FRAGMENT
)

ALL_GITHUB_TOKENS_WITH_ORG_ACCESS = (
    """
{
  allGitHubTokens(filters: { scopeIlike: "%read:org%" }) {
    totalCount
    edges {
      node {
        ...GitHubTokenFragment
      }
    }
  }
}
"""
    + GITHUB_TOKEN_FRAGMENT
)

HEADERS = {
    "Authorization": "Bearer token1"  # user1
}


##__________________________________________________________________||
params = [
    pytest.param(
        {"query": ALL_GITHUB_TOKENS},
        id="simple",
    ),
    pytest.param(
        {"query": ALL_GITHUB_TOKENS_WITH_ORG_ACCESS},
        id="filter",
    ),
]


##__________________________________________________________________||
@pytest.mark.parametrize("data", params)
@pytest.mark.asyncio
async def test_schema(app, snapshot, data):
    await assert_query(app, snapshot, data, HEADERS)

##__________________________________________________________________||



from ..fragments import FRAGMENT_PRODUCT_TYPE_CONNECTION

QUERY_ALL_PRODUCT_TYPES = (
    """
query AllProductTypes($sort: [ProductTypeSortEnum] = [ORDER_ASC]) {
  allProductTypes(sort: $sort) {
    ...fragmentProductTypeConnection
  }
}
"""
    + FRAGMENT_PRODUCT_TYPE_CONNECTION
)



import json

import pytest

from acondbs import ops


##__________________________________________________________________||
@pytest.fixture
def app(app_users):
    y = app_users

    config_json = json.dumps(
        {
            "headTitle": "Head Title",
            "toolbarTitle": "Toolbar Title",
            "devtoolLoadingstate": True,
            "productCreationDialog": False,
            "productUpdateDialog": True,
            "productDeletionDialog": True,
        },
        indent=2,
    )

    with y.app_context():
        ops.save_web_config(json=config_json)
        ops.commit()

    yield y


##__________________________________________________________________||



from acondbs.db.sa import sa
from acondbs.models import GitHubOrg


##__________________________________________________________________||
def test_entry(app_empty):
    app = app_empty

    row = GitHubOrg(
        login="urban-octo-disco", git_hub_id="012:Organization75631844"
    )

    with app.app_context():
        sa.session.add(row)
        sa.session.commit()

    with app.app_context():
        row = GitHubOrg.query.filter_by(login="urban-octo-disco").one()
        assert "urban-octo-disco" == row.login


##__________________________________________________________________||



"""rename the table github_admin_app_token github_tokens

Revision ID: bad578ddbf3a
Revises: 63033c01def0
Create Date: 2020-12-26 13:14:55.257552

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'bad578ddbf3a'
down_revision = '63033c01def0'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.rename_table('github_admin_app_token', 'github_tokens')
    # op.create_table('github_tokens',
    # sa.Column('token_id', sa.Integer(), nullable=False),
    # sa.Column('token', sqlalchemy_utils.types.encrypted.encrypted_type.EncryptedType(), nullable=True),
    # sa.Column('scope', sa.Text(), nullable=True),
    # sa.Column('user_id', sa.Integer(), nullable=False),
    # sa.ForeignKeyConstraint(['user_id'], ['github_users.user_id'], name=op.f('fk_github_tokens_user_id_github_users')),
    # sa.PrimaryKeyConstraint('token_id', name=op.f('pk_github_tokens'))
    # )
    # op.drop_table('github_admin_app_token')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.rename_table('github_tokens', 'github_admin_app_token')
    # op.create_table('github_admin_app_token',
    # sa.Column('token_id', sa.INTEGER(), nullable=False),
    # sa.Column('token', sa.BLOB(), nullable=True),
    # sa.Column('scope', sa.TEXT(), nullable=True),
    # sa.Column('user_id', sa.INTEGER(), nullable=False),
    # sa.ForeignKeyConstraint(['user_id'], ['github_users.user_id'], name='fk_github_admin_app_token_user_id_github_users'),
    # sa.PrimaryKeyConstraint('token_id', name='pk_admin_app_token')
    # )
    # op.drop_table('github_tokens')
    # ### end Alembic commands ###



from async_asgi_testclient import TestClient
from a2wsgi import WSGIMiddleware


HEADERS_DEFAULT = {"Content-Type:": "application/json"}


##__________________________________________________________________||
async def assert_query(app, snapshot, data, headers=None, error=False):
    if not headers:
        headers = {}

    app = WSGIMiddleware(app)  # convert a wsgi app to an asgi app

    headers_custom = headers
    headers = HEADERS_DEFAULT.copy()
    headers.update(headers_custom)

    async with TestClient(app) as client:
        resp = await client.post("/graphql", json=data, headers=headers)

    resp_json = resp.json()

    # assert errors
    if error:
        assert "errors" in resp_json
        return

    assert "errors" not in resp_json

    # snapshot test
    #   https://github.com/syrusakbary/snapshottest/
    snapshot.assert_match(resp_json)


async def assert_mutation(
    app,
    snapshot,
    data_mutation,
    headers_mutation,
    data_query,
    headers_query,
    mock_request_backup_db,
    success=True,
):
    if success:
        mock_request_backup_db.reset_mock()

    await assert_query(
        app, snapshot, data_mutation, headers_mutation, error=not success
    )

    if success:
        mock_request_backup_db.assert_called_once()

    await assert_query(app, snapshot, data_query, headers_query)


##__________________________________________________________________||



# -*- coding: utf-8 -*-
# snapshottest: v1 - https://goo.gl/zC4yUc
from __future__ import unicode_literals

from snapshottest import Snapshot


snapshots = Snapshot()

snapshots['test_schema_error[deleteProductFilePath-error] 1'] = {
    'data': {
        'allProductFilePaths': {
            'edges': [
                {
                    'node': {
                        'productId': 1001
                    }
                },
                {
                    'node': {
                        'productId': 1012
                    }
                },
                {
                    'node': {
                        'productId': 1012
                    }
                },
                {
                    'node': {
                        'productId': 1013
                    }
                },
                {
                    'node': {
                        'productId': 1070
                    }
                },
                {
                    'node': {
                        'productId': 1120
                    }
                },
                {
                    'node': {
                        'productId': 1130
                    }
                },
                {
                    'node': {
                        'productId': 1150
                    }
                },
                {
                    'node': {
                        'productId': 1002
                    }
                },
                {
                    'node': {
                        'productId': 1002
                    }
                }
            ]
        }
    }
}

snapshots['test_schema_success[deleteProductFilePath] 1'] = {
    'data': {
        'deleteProductFilePath': {
            'ok': True
        }
    }
}

snapshots['test_schema_success[deleteProductFilePath] 2'] = {
    'data': {
        'product': {
            'name': 'lat20190213',
            'note': '''- This is a dummy test with a lat map
- This should not depend on any beam''',
            'paths': {
                'edges': [
                ]
            },
            'producedBy': 'pwg-pmn',
            'timePosted': '2019-02-13T10:05:23'
        }
    }
}



from ..db.sa import sa
from ..models import ProductRelationType


##__________________________________________________________________||
def create_product_relation_type(type_, reverse=None, self_reverse=False):
    """Create a product relation type"""

    if self_reverse and reverse:
        raise ValueError('"reverse" is given when "self_reverse" is True')

    if not (self_reverse or reverse):
        raise ValueError('"reverse" is not given when "self_reverse" is False')

    model = ProductRelationType(**type_)
    if self_reverse:
        model.reverse = model
    else:
        model.reverse = ProductRelationType(**reverse)

    sa.session.add(model)
    return model


def update_product_relation_type(type_id, **kwargs):
    """Update a product relation type"""
    model = ProductRelationType.query.filter_by(type_id=type_id).one()
    for k, v in kwargs.items():
        setattr(model, k, v)
    return model


def delete_product_relation_type(type_id):
    """Delete a product relation type"""
    model = ProductRelationType.query.filter_by(type_id=type_id).one()
    sa.session.delete(model)


##__________________________________________________________________||



import graphene
from graphene import relay
from graphene_sqlalchemy import SQLAlchemyObjectType

from ..connection import CountedConnection
from ...models import (
    GitHubUser as GitHubUserModel,
    GitHubOrg as GitHubOrgModel,
    GitHubOrgMembership as GitHubOrgMembershipModel,
    GitHubToken as GitHubTokenModel,
)
from ..filter_ import PFilterableConnectionField


##__________________________________________________________________||
class GitHubOAuthAppInfo(graphene.ObjectType):
    client_id = graphene.String()
    authorize_url = graphene.String()
    redirect_uri = graphene.String()


##__________________________________________________________________||
class AuthPayload(graphene.ObjectType):
    token = graphene.String()


##__________________________________________________________________||
class GitHubToken(SQLAlchemyObjectType):
    class Meta:
        model = GitHubTokenModel
        interfaces = (relay.Node,)
        exclude_fields = [
            "token",
        ]
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory

    token_masked = graphene.String()

    def resolve_token_masked(parent, info):
        return "X" * 15


##__________________________________________________________________||
class GitHubUser(SQLAlchemyObjectType):
    class Meta:
        model = GitHubUserModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


##__________________________________________________________________||
class GitHubOrg(SQLAlchemyObjectType):
    class Meta:
        model = GitHubOrgModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


##__________________________________________________________________||
class GitHubOrgMembership(SQLAlchemyObjectType):
    class Meta:
        model = GitHubOrgMembershipModel
        interfaces = (relay.Node,)
        connection_class = CountedConnection
        connection_field_factory = PFilterableConnectionField.factory


##__________________________________________________________________||



"""remove table web_config_old

Revision ID: e69be63fc3b8
Revises: 1f86dd56d829
Create Date: 2021-11-02 11:44:47.061737

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e69be63fc3b8'
down_revision = '1f86dd56d829'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('web_config_old')
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('web_config_old',
    sa.Column('config_id', sa.INTEGER(), nullable=False),
    sa.Column('head_title', sa.TEXT(), nullable=True),
    sa.Column('toolbar_title', sa.TEXT(), nullable=True),
    sa.Column('devtool_loadingstate', sa.BOOLEAN(), nullable=True),
    sa.Column('product_creation_dialog', sa.BOOLEAN(), nullable=True),
    sa.Column('product_update_dialog', sa.BOOLEAN(), nullable=True),
    sa.Column('product_deletion_dialog', sa.BOOLEAN(), nullable=True),
    sa.PrimaryKeyConstraint('config_id', name='pk_web_config')
    )
    # ### end Alembic commands ###



from ..fragments import FRAGMENT_PRODUCT_TYPE

MUTATION_UPDATE_PRODUCT_TYPE = '''
mutation UpdateProductType($typeId: Int!, $input: UpdateProductTypeInput!) {
  updateProductType(typeId: $typeId,input: $input) {
    ok
    productType {
      ...fragmentProductType
    }
  }
}
''' + FRAGMENT_PRODUCT_TYPE




from ..fragments import FRAGMENT_PRODUCT_RELATION_CONNECTION

QUERY_ALL_PRODUCT_RELATIONS = (
    """
{
  allProductRelations {
    ...fragmentProductRelationConnection
  }
}
"""
    + FRAGMENT_PRODUCT_RELATION_CONNECTION
)



import pytest

import datetime

from acondbs.db.sa import sa
from acondbs.models import (
    ProductType,
    FieldType,
    Field,
    TypeFieldAssociation,
    Product,
)


##__________________________________________________________________||
@pytest.fixture
def app(app_empty):

    y = app_empty

    # create fields
    field1 = Field(name="attr1", type_=FieldType.UnicodeText)
    field2 = Field(name="date_produced", type_=FieldType.Date)
    field3 = Field(name="time_posted", type_=FieldType.DateTime)
    fields = [field1, field2, field3]

    # create product types
    Map = ProductType(
        type_id=1,
        name="map",
        order=2,
        indef_article="a",
        singular="map",
        plural="maps",
        icon="mdi-map",
        fields=[TypeFieldAssociation(field=f) for f in fields],
    )
    Beam = ProductType(
        type_id=2,
        name="beam",
        order=1,
        indef_article="a",
        singular="beam",
        plural="beams",
        icon="mdi-spotlight-beam",
        fields=[TypeFieldAssociation(field=f) for f in fields],
    )

    # create products
    map1 = Product(product_id=1, name="map1", type_=Map)
    values = (
        "value1",
        datetime.date(2020, 2, 1),
        datetime.datetime(2020, 2, 1, 9, 10, 25),
    )
    for assoc, value in zip(Map.fields, values):
        assoc.field.type_.attribute_class(
            product=map1,
            type_field_association=assoc,
            field=assoc.field,
            value=value,
        )

    map2 = Product(product_id=2, name="map2", type_=Map)
    values = (
        "value2",
        datetime.date(2020, 2, 10),
        datetime.datetime(2020, 2, 10, 13, 20, 2),
    )
    for assoc, value in zip(Map.fields, values):
        assoc.field.type_.attribute_class(
            product=map2,
            type_field_association=assoc,
            field=assoc.field,
            value=value,
        )

    map3 = Product(product_id=3, name="map3", type_=Map)
    values = (
        "value3",
        datetime.date(2020, 3, 19),
        datetime.datetime(2020, 3, 20, 8, 45, 30),
    )
    for assoc, value in zip(Map.fields, values):
        assoc.field.type_.attribute_class(
            product=map3,
            type_field_association=assoc,
            field=assoc.field,
            value=value,
        )

    beam1 = Product(product_id=4, name="beam1", type_=Beam)
    values = (
        "value4",
        datetime.date(2020, 2, 5),
        datetime.datetime(2020, 2, 5, 12, 3, 48),
    )
    for assoc, value in zip(Map.fields, values):
        assoc.field.type_.attribute_class(
            product=beam1,
            type_field_association=assoc,
            field=assoc.field,
            value=value,
        )

    beam2 = Product(product_id=5, name="beam2", type_=Beam)
    values = (
        "value5",
        datetime.date(2020, 3, 4),
        datetime.datetime(2020, 3, 4, 19, 22, 5),
    )
    for assoc, value in zip(Map.fields, values):
        assoc.field.type_.attribute_class(
            product=beam2,
            type_field_association=assoc,
            field=assoc.field,
            value=value,
        )

    with y.app_context():
        sa.session.add(Map)
        sa.session.add(Beam)
        sa.session.commit()

    yield y


##__________________________________________________________________||



"""update type_field_association

Revision ID: f84493df4297
Revises: 6bd27755b1a8
Create Date: 2021-11-02 16:07:32.303654

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f84493df4297'
down_revision = '6bd27755b1a8'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('type_field_association', schema=None) as batch_op:
        batch_op.add_column(sa.Column('iid', sa.Integer(), nullable=False))
        batch_op.alter_column('type_id',
               existing_type=sa.INTEGER(),
               nullable=True)
        batch_op.drop_constraint('pk_type_field_association', type_='primary')
        batch_op.create_unique_constraint('_type_field', ['type_id', 'field_id'])
        batch_op.create_primary_key('pk_type_field_association', ['iid'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('type_field_association', schema=None) as batch_op:
        batch_op.drop_constraint('_type_field', type_='unique')
        batch_op.alter_column('type_id',
               existing_type=sa.INTEGER(),
               nullable=False)
        batch_op.drop_column('iid')
        batch_op.drop_constraint('pk_type_field_association', type_='primary')
        batch_op.create_primary_key('pk_type_field_association', ['type_id', 'field_id'])

    # ### end Alembic commands ###



from ..fragments import FRAGMENT_PRODUCT_RELATION_TYPE

MUTATION_CREATE_PRODUCT_RELATION_TYPES = '''
mutation CreateProductRelationTypes(
  $type: CreateProductRelationTypeInput!
  $reverse: CreateProductRelationTypeInput
  $selfReverse: Boolean
) {
  createProductRelationTypes(
    type: $type
    reverse: $reverse
    selfReverse: $selfReverse
  ) {
    ok
    productRelationType {
      ...fragmentProductRelationType
    }
  }
}
''' + FRAGMENT_PRODUCT_RELATION_TYPE



from acondbs.db.sa import sa
from acondbs.models import AccountAdmin


##__________________________________________________________________||
def test_entry(app_empty):
    app = app_empty

    admin1 = AccountAdmin(git_hub_login="octocat")

    with app.app_context():
        sa.session.add(admin1)
        sa.session.commit()

    with app.app_context():
        admin1 = AccountAdmin.query.filter_by(git_hub_login="octocat").one()
        assert "octocat" == admin1.git_hub_login


##__________________________________________________________________||



import json

import pytest

from ....funcs import assert_mutation

from ...gql import MUTATION_SAVE_WEB_CONFIG, QUERY_WEB_CONFIG

HEADERS_MUTATION = {
    "Authorization": "Bearer 90b2ee5fed25506df04fd37343bb68d1803dd97f"  # octocat
}

HEADERS_QUERY = {}

CONFIG_JSON = json.dumps(
    {
        "headTitle": "Saved Head Title",
        "toolbarTitle": "Saved Toolbar Title",
        "devtoolLoadingstate": False,
        "productCreationDialog": False,
        "productUpdateDialog": False,
        "productDeletionDialog": False,
    },
    indent=2,
)


##__________________________________________________________________||
params = [
    pytest.param(
        {
            "query": MUTATION_SAVE_WEB_CONFIG,
            "variables": {
                "json": CONFIG_JSON,
            },
        },
        {"query": QUERY_WEB_CONFIG},
        id="one",
    ),
]


@pytest.mark.parametrize("data_mutation, data_query", params)
@pytest.mark.asyncio
async def test_schema_success(
    app, snapshot, data_mutation, data_query, mock_request_backup_db
):

    success = True
    await assert_mutation(
        app,
        snapshot,
        data_mutation,
        HEADERS_MUTATION,
        data_query,
        HEADERS_QUERY,
        mock_request_backup_db,
        success,
    )


##__________________________________________________________________||



import pytest

from acondbs.db.sa import sa
from acondbs.models import ProductRelationType


##__________________________________________________________________||
@pytest.fixture
def app(app_empty):

    y = app_empty

    #
    #  +--------+                +-------+
    #  |        | --(reverse)->  |       |
    #  | parent |                | child |
    #  |        | <-(reverse)--  |       |
    #  +--------+                +-------+
    #

    #
    #  +------- -+
    #  |         | --(reverse)-+
    #  | sibling |             |
    #  |         | <-----------+
    #  +------ --+
    #

    parent = ProductRelationType(name="parent")
    child = ProductRelationType(name="child")
    parent.reverse = child
    assert child.reverse == parent

    sibling = ProductRelationType(name="sibling")
    sibling.reverse = sibling

    # commit
    with y.app_context():
        sa.session.add(parent)
        sa.session.add(sibling)
        sa.session.commit()
    yield y


##__________________________________________________________________||
def test_reverse(app):

    with app.app_context():
        parent = ProductRelationType.query.filter_by(name="parent").one()
        child = ProductRelationType.query.filter_by(name="child").one()

        assert child is parent.reverse
        assert parent is child.reverse


def test_self_reverse(app):

    with app.app_context():
        sibling = ProductRelationType.query.filter_by(name="sibling").one()
        assert sibling is sibling.reverse


def test_cascade(app):

    # delete parent
    with app.app_context():
        parent = ProductRelationType.query.filter_by(name="parent").one()
        sa.session.delete(parent)
        sa.session.commit()

    # assert
    with app.app_context():
        parent = ProductRelationType.query.filter_by(
            name="parent"
        ).one_or_none()
        child = ProductRelationType.query.filter_by(name="child").one_or_none()

        assert parent is None
        assert child is None


##__________________________________________________________________||



"""add column to attribute tables

Revision ID: e58cc402c887
Revises: 0ff075f16b94
Create Date: 2021-11-03 16:32:18.594542

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'e58cc402c887'
down_revision = '0ff075f16b94'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('attribute_boolean', schema=None) as batch_op:
        batch_op.add_column(sa.Column('type_field_association_iid', sa.Integer(), nullable=True))
        batch_op.create_foreign_key(batch_op.f('fk_attribute_boolean_type_field_association_iid_type_field_association'), 'type_field_association', ['type_field_association_iid'], ['iid'])

    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.add_column(sa.Column('type_field_association_iid', sa.Integer(), nullable=True))
        batch_op.create_foreign_key(batch_op.f('fk_attribute_date_type_field_association_iid_type_field_association'), 'type_field_association', ['type_field_association_iid'], ['iid'])

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.add_column(sa.Column('type_field_association_iid', sa.Integer(), nullable=True))
        batch_op.create_foreign_key(batch_op.f('fk_attribute_date_time_type_field_association_iid_type_field_association'), 'type_field_association', ['type_field_association_iid'], ['iid'])

    with op.batch_alter_table('attribute_float', schema=None) as batch_op:
        batch_op.add_column(sa.Column('type_field_association_iid', sa.Integer(), nullable=True))
        batch_op.create_foreign_key(batch_op.f('fk_attribute_float_type_field_association_iid_type_field_association'), 'type_field_association', ['type_field_association_iid'], ['iid'])

    with op.batch_alter_table('attribute_integer', schema=None) as batch_op:
        batch_op.add_column(sa.Column('type_field_association_iid', sa.Integer(), nullable=True))
        batch_op.create_foreign_key(batch_op.f('fk_attribute_integer_type_field_association_iid_type_field_association'), 'type_field_association', ['type_field_association_iid'], ['iid'])

    with op.batch_alter_table('attribute_time', schema=None) as batch_op:
        batch_op.add_column(sa.Column('type_field_association_iid', sa.Integer(), nullable=True))
        batch_op.create_foreign_key(batch_op.f('fk_attribute_time_type_field_association_iid_type_field_association'), 'type_field_association', ['type_field_association_iid'], ['iid'])

    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.add_column(sa.Column('type_field_association_iid', sa.Integer(), nullable=True))
        batch_op.create_foreign_key(batch_op.f('fk_attribute_unicode_text_type_field_association_iid_type_field_association'), 'type_field_association', ['type_field_association_iid'], ['iid'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('attribute_unicode_text', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_unicode_text_type_field_association_iid_type_field_association'), type_='foreignkey')
        batch_op.drop_column('type_field_association_iid')

    with op.batch_alter_table('attribute_time', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_time_type_field_association_iid_type_field_association'), type_='foreignkey')
        batch_op.drop_column('type_field_association_iid')

    with op.batch_alter_table('attribute_integer', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_integer_type_field_association_iid_type_field_association'), type_='foreignkey')
        batch_op.drop_column('type_field_association_iid')

    with op.batch_alter_table('attribute_float', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_float_type_field_association_iid_type_field_association'), type_='foreignkey')
        batch_op.drop_column('type_field_association_iid')

    with op.batch_alter_table('attribute_date_time', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_date_time_type_field_association_iid_type_field_association'), type_='foreignkey')
        batch_op.drop_column('type_field_association_iid')

    with op.batch_alter_table('attribute_date', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_date_type_field_association_iid_type_field_association'), type_='foreignkey')
        batch_op.drop_column('type_field_association_iid')

    with op.batch_alter_table('attribute_boolean', schema=None) as batch_op:
        batch_op.drop_constraint(batch_op.f('fk_attribute_boolean_type_field_association_iid_type_field_association'), type_='foreignkey')
        batch_op.drop_column('type_field_association_iid')

    # ### end Alembic commands ###



from flask import request

from ..models import (
    GitHubToken,
    GitHubUser,
    AccountAdmin
)

##__________________________________________________________________||
def is_signed_in():
    """
    """
    token = _get_token_from_http_headers()
    if not token:
        return False

    token_model = GitHubToken.query.filter_by(token=token).one_or_none()
    if not token_model:
        return False

    return True

##__________________________________________________________________||
def is_admin():
    """
    """

    if not is_signed_in():
        return False

    token = _get_token_from_http_headers()
    if not token:
        return False

    user_model = GitHubUser.query.join(GitHubToken). \
        filter(GitHubToken.token==token). \
        one_or_none()

    if not user_model:
        return False

    admin_model = AccountAdmin.query. \
        filter_by(git_hub_login=user_model.login). \
        one_or_none()

    if not admin_model:
        return False

    return True

##__________________________________________________________________||
def _get_token_from_http_headers():
    auth_header = request.headers.get('Authorization')
    # e.g., 'Bearer "xxxx"', "Bearer 'xxxx'",  or 'Bearer xxxx'

    if not auth_header:
        return None

    token = auth_header.split()[1].strip('"\'')
    # e.g., "xxxx"

    return token

##__________________________________________________________________||
